@@ -3,12 +3,10 @@ package machine
 import (
 	"context"
 	"fmt"
-	"strings"
 
 	"github.com/spf13/cobra"
 	"github.com/superfly/flyctl/api"
 	"github.com/superfly/flyctl/flaps"
-	"github.com/superfly/flyctl/internal/appconfig"
 	"github.com/superfly/flyctl/internal/command"
 	"github.com/superfly/flyctl/internal/flag"
 	"github.com/superfly/flyctl/iostreams"
@@ -44,7 +42,12 @@ func runMachineStop(ctx context.Context) (err error) {
 		args = flag.Args(ctx)
 	)
 
-	for _, machineID := range args {
+	machineIDs, ctx, err := selectManyMachineIDs(ctx, args)
+	if err != nil {
+		return err
+	}
+
+	for _, machineID := range machineIDs {
 		fmt.Fprintf(io.Out, "Sending kill signal to machine %s...\n", machineID)
 
 		if err = Stop(ctx, machineID); err != nil {
@@ -56,33 +59,17 @@ func runMachineStop(ctx context.Context) (err error) {
 }
 
 func Stop(ctx context.Context, machineID string) (err error) {
-	var (
-		appName = appconfig.NameFromContext(ctx)
-	)
-
 	machineStopInput := api.StopMachineInput{
 		ID:      machineID,
 		Filters: &api.Filters{},
 	}
 
-	app, err := appFromMachineOrName(ctx, machineID, appName)
+	err = flaps.FromContext(ctx).Stop(ctx, machineStopInput)
 	if err != nil {
-		return fmt.Errorf("could not get app: %w", err)
-	}
-
-	flapsClient, err := flaps.New(ctx, app)
-	if err != nil {
-		return fmt.Errorf("could not make flaps client: %w", err)
-	}
-
-	err = flapsClient.Stop(ctx, machineStopInput)
-	if err != nil {
-		switch {
-		case strings.Contains(err.Error(), "not found") && appName != "":
-			return fmt.Errorf("machine %s was not found in app %s", machineID, appName)
-		default:
-			return fmt.Errorf("could not stop machine %s: %w", machineStopInput.ID, err)
+		if err := rewriteMachineNotFoundErrors(ctx, err, machineID); err != nil {
+			return err
 		}
+		return fmt.Errorf("could not stop machine %s: %w", machineID, err)
 	}
 
 	return
