@@ -6,6 +6,7 @@ import (
 	"fmt"
 
 	"github.com/superfly/flyctl/api"
+	"github.com/superfly/flyctl/internal/appconfig"
 	"github.com/superfly/flyctl/internal/command"
 	"github.com/superfly/flyctl/internal/command/deploy"
 	"github.com/superfly/flyctl/internal/flag"
@@ -67,7 +68,14 @@ func deployForSecrets(ctx context.Context, app *api.AppCompact, release *api.Rel
 	}
 
 	if app.PlatformVersion == "machines" {
-		ctx, err = command.LoadAppConfigIfPresent(ctx)
+		// It would be confusing for setting secrets to deploy the current fly.toml file.
+		// Instead, we always grab the currently deployed app config
+		cfg, err := appconfig.FromRemoteApp(ctx, app.Name)
+		if err != nil {
+			return err
+		}
+		ctx = appconfig.WithConfig(ctx, cfg)
+
 		if err != nil {
 			return fmt.Errorf("error loading appv2 config: %w", err)
 		}
